import pydantic.version

if not pydantic.version.VERSION.startswith("1"):  # pragma: no cover
    raise ImportError("pydantic_compat._v1 only supports pydantic v1.x")


from .decorators import field_validator as field_validator
from .decorators import model_validator as model_validator
from .decorators import root_validator as root_validator
from .mixin import PydanticCompatMixin as PydanticCompatMixin



import contextlib
import warnings
from typing import Any

import pydantic
import pydantic.version

PYDANTIC2 = pydantic.version.VERSION.startswith("2")

V2_REMOVED_CONFIG_KEYS = {
    "allow_mutation",
    "error_msg_templates",
    "fields",
    "getter_dict",
    "smart_union",
    "underscore_attrs_are_private",
    "json_loads",
    "json_dumps",
    "copy_on_model_validation",
    "post_init_call",
}
V2_RENAMED_CONFIG_KEYS = {
    "allow_population_by_field_name": "populate_by_name",
    "anystr_lower": "str_to_lower",
    "anystr_strip_whitespace": "str_strip_whitespace",
    "anystr_upper": "str_to_upper",
    "keep_untouched": "ignored_types",
    "max_anystr_length": "str_max_length",
    "min_anystr_length": "str_min_length",
    "orm_mode": "from_attributes",
    "schema_extra": "json_schema_extra",
    "validate_all": "validate_default",
}

V1_FIELDS_TO_V2_FIELDS = {
    "min_items": "min_length",
    "max_items": "max_length",
    "regex": "pattern",
    "allow_mutation": "-frozen",
    "metadata": "json_schema_extra",
}

V2_FIELDS_TO_V1_FIELDS = {}
for k, v in V1_FIELDS_TO_V2_FIELDS.items():
    if v.startswith("-"):
        v = v[1:]
        k = f"-{k}"
    V2_FIELDS_TO_V1_FIELDS[v] = k

FIELD_NAME_MAP = V1_FIELDS_TO_V2_FIELDS if PYDANTIC2 else V2_FIELDS_TO_V1_FIELDS


def check_mixin_order(cls: type, mixin_class: type, base_model: type) -> None:
    """Warn if mixin_class appears after base_model in cls.__bases__."""
    bases = cls.__bases__
    with contextlib.suppress(ValueError):
        mixin_index = bases.index(mixin_class)
        base_model_index = bases.index(base_model)
        if mixin_index > base_model_index:
            warnings.warn(
                f"{mixin_class.__name__} should appear before pydantic.BaseModel",
                stacklevel=3,
            )


def move_field_kwargs(kwargs: dict) -> dict:
    """Move Field(...) kwargs from v1 to v2 and vice versa."""
    for old_name, new_name in FIELD_NAME_MAP.items():
        negate = False
        if new_name.startswith("-"):
            new_name = new_name[1:]
            negate = True
        if old_name in kwargs:
            if new_name in kwargs:
                raise ValueError(f"Cannot specify both {old_name} and {new_name}")
            val = not kwargs.pop(old_name) if negate else kwargs.pop(old_name)
            kwargs[new_name] = val
    return kwargs


def clean_field_kwargs(kwargs: dict) -> dict:
    """Remove outdated Field(...) kwargs."""
    const = kwargs.pop("const", None)
    if const is not None:
        raise TypeError(
            f"`const` is removed in v2, use `Literal[{const!r}]` instead, "
            "it works in v1 and v2."
        )
    return kwargs


def Field(*args: Any, **kwargs: Any) -> Any:
    """Create a field for objects that can be configured."""
    kwargs = clean_field_kwargs(kwargs)
    kwargs = move_field_kwargs(kwargs)
    return pydantic.Field(*args, **kwargs)



from __future__ import annotations

import warnings
from typing import Any, Callable

from pydantic.deprecated import class_validators


# V1 signature
# def validator(
#     *fields: str,
#     pre: bool = False,
#     each_item: bool = False,
#     always: bool = False,
#     check_fields: bool = True,
#     whole: Optional[bool] = None,
#     allow_reuse: bool = False,
# ) -> Callable[[AnyCallable], 'AnyClassMethod']:
#     ...
def validator(
    _field: str, *fields: str, **kwargs: Any
) -> Callable[[Callable], Callable]:
    """Adaptor from v1.validator -> v2.field_validator."""
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", DeprecationWarning)
        return class_validators.validator(_field, *fields, **kwargs)


# V1 signature
# def root_validator(
#     _func: Optional[AnyCallable] = None,
#     *,
#     pre: bool = False,
#     allow_reuse: bool = False,
#     skip_on_failure: bool = False,
# ) -> Union["AnyClassMethod", Callable[[AnyCallable], "AnyClassMethod"]]:
#     ...
def root_validator(
    *_args: str,
    pre: bool = False,
    skip_on_failure: bool | None = None,
    allow_reuse: bool = False,
    construct_object: bool = False,  # here to match our v1 patch behavior
) -> Any:
    """Adaptor from v1.root_validator -> v2.model_validator."""
    # If you use `@root_validator` with pre=False (the default)
    # you MUST specify `skip_on_failure=True`
    # we let explicit `skip_on_failure=False` pass through to fail,
    # but we default to `skip_on_failure=True` to match v1 behavior
    if not pre and skip_on_failure is None:
        skip_on_failure = True

    if construct_object:
        raise ValueError(
            "construct_object=True is not supported by pydantic-compat when running on "
            "pydantic v2. Please use pydantic_compat.model_validator(mode='after') "
            "instead. (It works for both versions)."
        )

    with warnings.catch_warnings():
        warnings.simplefilter("ignore", DeprecationWarning)

        # def model_validator( *, mode: Literal['wrap', 'before', 'after']) -> Any:
        return class_validators.root_validator(  # type: ignore [call-overload]
            *_args,
            pre=pre,
            skip_on_failure=bool(skip_on_failure),
            allow_reuse=allow_reuse,
        )



"""CompatibilityMixin for pydantic v1/1/v2."""

try:
    from importlib.metadata import PackageNotFoundError, version
except ImportError:
    from importlib_metadata import PackageNotFoundError, version  # type: ignore

from typing import TYPE_CHECKING

try:
    __version__ = version("pydantic-compat")
except PackageNotFoundError:  # pragma: no cover
    __version__ = "uninstalled"

__author__ = "Talley Lambert"
__email__ = "talley.lambert@gmail.com"
__all__ = [
    "__version__",
    "field_validator",
    "model_validator",
    "PYDANTIC2",
    "PydanticCompatMixin",
    "root_validator",
    "validator",
    "Field",
    "BaseModel",
]

from ._shared import PYDANTIC2

if TYPE_CHECKING:
    from pydantic import (
        Field,
        field_validator,
        model_validator,
        root_validator,
        validator,
    )

    # using this to avoid breaking pydantic mypy plugin
    # not that we could use a protocol. but it will be hard to provide proper names
    # AND proper signatures for both versions of pydantic without a ton of potentially
    # outdated signatures
    PydanticCompatMixin = type
else:
    from ._shared import Field

    if PYDANTIC2:
        from pydantic import field_validator, model_validator

        from ._v2 import PydanticCompatMixin, root_validator, validator

    else:
        from pydantic import validator

        from ._v1 import (
            PydanticCompatMixin,
            field_validator,
            model_validator,
            root_validator,
        )


import pydantic


class BaseModel(PydanticCompatMixin, pydantic.BaseModel):
    """BaseModel with pydantic_compat mixins."""


del pydantic
